from __future__ import annotations
from dataclasses import dataclass, field, fields
from typing import Optional, Any, TYPE_CHECKING
import logging

from accqsure.exceptions import SpecificationError

if TYPE_CHECKING:
    from accqsure import AccQsure


class Inspections(object):
    def __init__(self, accqsure):

        self.accqsure = accqsure

    async def get(self, id_, **kwargs):

        resp = await self.accqsure._query(f"/inspection/{id_}", "GET", kwargs)
        return Inspection.from_api(self.accqsure, resp)

    async def list(self, inspection_type, limit=50, start_key=None, **kwargs):

        resp = await self.accqsure._query(
            "/inspection",
            "GET",
            {
                "type": inspection_type,
                "limit": limit,
                "start_key": start_key,
                **kwargs,
            },
        )
        inspections = [
            Inspection.from_api(self.accqsure, inspection)
            for inspection in resp.get("results")
        ]
        return inspections, resp.get("last_key")

    async def create(
        self,
        inspection_type,
        name,
        document_type_id,
        manifests,
        draft=None,
        documents=None,
        **kwargs,
    ):

        data = dict(
            name=name,
            type=inspection_type,
            document_type_id=document_type_id,
            manifests=manifests,
            draft=draft,
            documents=documents,
            **kwargs,
        )
        payload = {k: v for k, v in data.items() if v is not None}
        logging.info("Creating Inspection %s", name)

        resp = await self.accqsure._query("/inspection", "POST", None, payload)
        inspection = Inspection.from_api(self.accqsure, resp)
        logging.info("Created Inspection %s with id %s", name, inspection.id)

        return inspection

    async def remove(self, id_, **kwargs):

        await self.accqsure._query(f"/inspection/{id_}", "DELETE", {**kwargs})


@dataclass
class Inspection:
    id: str
    name: str
    type: str
    status: str
    created_at: Optional[str] = field(default=None)
    updated_at: Optional[str] = field(default=None)
    document_type_id: Optional[str] = field(default=None)
    doc_content_id: Optional[str] = field(default=None)
    content_id: Optional[str] = field(default=None)

    @classmethod
    def from_api(
        cls, accqsure: "AccQsure", data: dict[str, Any]
    ) -> "Inspection":
        if not data:
            return None
        entity = cls(
            id=data.get("entity_id"),
            name=data.get("name"),
            type=data.get("type"),
            status=data.get("status"),
            created_at=data.get("created_at"),
            updated_at=data.get("updated_at"),
            document_type_id=data.get("document_type_id"),
            doc_content_id=data.get("doc_content_id"),
            content_id=data.get("content_id"),
        )
        entity.accqsure = accqsure
        return entity

    @property
    def accqsure(self) -> "AccQsure":
        return self._accqsure

    @accqsure.setter
    def accqsure(self, value: "AccQsure"):
        self._accqsure = value

    async def remove(self):

        await self.accqsure._query(
            f"/inspection/{self.id}",
            "DELETE",
        )

    async def rename(self, name):

        resp = await self.accqsure._query(
            f"/inspection/{self.id}",
            "PUT",
            None,
            dict(name=name),
        )
        exclude = ["id", "accqsure"]

        for f in fields(self.__class__):
            if (
                f.name not in exclude and f.init and resp.get(f.name) is not None
            ):  # Only update init args
                setattr(self, f.name, resp.get(f.name))
        return self

    async def run(self):

        resp = await self.accqsure._query(
            f"/inspection/{self.id}/run",
            "POST",
        )
        exclude = ["id", "accqsure"]

        for f in fields(self.__class__):
            if (
                f.name not in exclude and f.init and resp.get(f.name) is not None
            ):  # Only update init args
                setattr(self, f.name, resp.get(f.name))
        return self

    async def refresh(self):

        resp = await self.accqsure._query(
            f"/inspection/{self.id}",
            "GET",
        )
        exclude = ["id", "accqsure"]

        for f in fields(self.__class__):
            if (
                f.name not in exclude and f.init and resp.get(f.name) is not None
            ):  # Only update init args
                setattr(self, f.name, resp.get(f.name))
        return self

    async def _set_asset(self, path, file_name, mime_type, contents):
        return await self.accqsure._query(
            f"/inspection/{self.id}/asset/{path}",
            "PUT",
            params={"file_name": file_name},
            data=contents,
            headers={"Content-Type": mime_type},
        )

    async def get_doc_contents(self):
        if not self.doc_content_id:
            raise SpecificationError(
                "doc_content_id",
                "Document content not uploaded for inspection",
            )

        resp = await self.accqsure._query(
            f"/inspection/{self.id}/asset/{self.doc_content_id}/manifest.json",
            "GET",
        )
        return resp

    async def get_doc_content_item(self, name):
        if not self.doc_content_id:
            raise SpecificationError(
                "doc_content_id", "Document not uploaded for inspection"
            )

        return await self.accqsure._query(
            f"/inspection/{self.id}/asset/{self.doc_content_id}/{name}",
            "GET",
        )

    async def _set_doc_content_item(
        self, name, file_name, mime_type, contents
    ):
        if not self.doc_content_id:
            raise SpecificationError(
                "content_id", "Content not finalized for inspection"
            )
        return await self._set_asset(
            f"{self.doc_content_id}/{name}", file_name, mime_type, contents
        )

    async def get_contents(self):
        if not self.content_id:
            raise SpecificationError(
                "content_id", "Content not finalized for inspection"
            )

        resp = await self.accqsure._query(
            f"/inspection/{self.id}/asset/{self.content_id}/manifest.json",
            "GET",
        )
        return resp

    async def get_content_item(self, name):
        if not self.content_id:
            raise SpecificationError(
                "content_id", "Content not finalized for inspection"
            )

        return await self.accqsure._query(
            f"/inspection/{self.id}/asset/{self.content_id}/{name}",
            "GET",
        )

    async def _set_content_item(self, name, file_name, mime_type, contents):
        if not self.content_id:
            raise SpecificationError(
                "content_id", "Content not finalized for inspection"
            )
        return await self._set_asset(
            f"{self.content_id}/{name}", file_name, mime_type, contents
        )

    async def download_report(self):
        if not self.content_id:
            raise SpecificationError(
                "content_id", "Content not finalized for inspection"
            )
        manifest = await self.get_contents()
        return await self.get_content_item(manifest.get("report"))

    async def list_checks(
        self,
        document_id=None,
        manifest_id=None,
        limit=50,
        start_key=None,
        name=None,
        **kwargs,
    ):

        resp = await self.accqsure._query(
            f"/inspection/{self.id}/check",
            "GET",
            {
                "document_id": document_id,
                "manifest_id": manifest_id,
                "limit": limit,
                "start_key": start_key,
                "name": name,
                **kwargs,
            },
        )
        checks = [
            InspectionCheck.from_api(self.accqsure, self.id, check)
            for check in resp.get("results")
        ]
        return checks, resp.get("last_key")


@dataclass
class InspectionCheck:
    inspection_id: str
    id: str
    section: str
    name: str
    status: str
    critical: Optional[bool] = field(default=None)
    compliant: Optional[bool] = field(default=None)
    rationale: Optional[str] = field(default=None)
    suggestion: Optional[str] = field(default=None)
    created_at: Optional[str] = field(default=None)
    updated_at: Optional[str] = field(default=None)

    @classmethod
    def from_api(
        cls, accqsure: "AccQsure", inspection_id: str, data: dict[str, Any]
    ) -> "InspectionCheck":
        if not data:
            return None
        entity = cls(
            inspection_id=inspection_id,
            id=data.get("entity_id"),
            section=data.get("check_section"),
            name=data.get("check_name"),
            status=data.get("status"),
            critical=data.get("critical"),
            compliant=data.get("compliant"),
            rationale=data.get("rationale"),
            suggestion=data.get("suggestion"),
            created_at=data.get("created_at"),
            updated_at=data.get("updated_at"),
        )
        entity.accqsure = accqsure
        return entity

    @property
    def accqsure(self) -> "AccQsure":
        return self._accqsure

    @accqsure.setter
    def accqsure(self, value: "AccQsure"):
        self._accqsure = value

    async def update(self, **kwargs):

        resp = await self.accqsure._query(
            f"/inspection/{self.inspection_id}/check/{self.id}",
            "PUT",
            None,
            dict(**kwargs),
        )
        exclude = ["id", "inspection_id", "accqsure"]

        for f in fields(self.__class__):
            if (
                f.name not in exclude and f.init and resp.get(f.name) is not None
            ):  # Only update init args
                # Handle field name mapping
                field_name = f.name
                if field_name == "section":
                    setattr(self, field_name, resp.get("check_section"))
                elif field_name == "name":
                    setattr(self, field_name, resp.get("check_name"))
                else:
                    setattr(self, field_name, resp.get(field_name))
        return self

    async def refresh(self):

        resp = await self.accqsure._query(
            f"/inspection/{self.inspection_id}/check/{self.id}",
            "GET",
        )
        exclude = ["id", "inspection_id", "accqsure"]

        for f in fields(self.__class__):
            if (
                f.name not in exclude and f.init and resp.get(f.name) is not None
            ):  # Only update init args
                # Handle field name mapping
                field_name = f.name
                if field_name == "section":
                    setattr(self, field_name, resp.get("check_section"))
                elif field_name == "name":
                    setattr(self, field_name, resp.get("check_name"))
                else:
                    setattr(self, field_name, resp.get(field_name))
        return self



from __future__ import annotations
from dataclasses import dataclass, field, fields
from typing import Optional, Any, TYPE_CHECKING
import logging

if TYPE_CHECKING:
    from accqsure import AccQsure


class DocumentTypes(object):
    def __init__(self, accqsure):
        self.accqsure = accqsure

    async def get(self, id_, **kwargs):
        resp = await self.accqsure._query(
            f"/document/type/{id_}", "GET", kwargs
        )
        return DocumentType.from_api(self.accqsure, resp)

    async def list(self, **kwargs):
        resp = await self.accqsure._query("/document/type", "GET", kwargs)
        document_types = [
            DocumentType.from_api(self.accqsure, document_type)
            for document_type in resp
        ]
        return document_types

    async def create(
        self,
        name,
        code,
        level,
        **kwargs,
    ):

        data = dict(
            name=name,
            code=code,
            level=level,
            **kwargs,
        )
        payload = {k: v for k, v in data.items() if v is not None}
        logging.info("Creating Document Type %s", name)
        resp = await self.accqsure._query(
            "/document/type", "POST", None, payload
        )
        document_type = DocumentType.from_api(self.accqsure, resp)
        logging.info(
            "Created Document Type %s with id %s", name, document_type.id
        )

        return document_type

    async def remove(self, id_, **kwargs):
        await self.accqsure._query(
            f"/document/type/{id_}", "DELETE", dict(**kwargs)
        )


@dataclass
class DocumentType:
    id: str
    name: str
    code: str
    level: int
    created_at: Optional[str] = field(default=None)
    updated_at: Optional[str] = field(default=None)

    @classmethod
    def from_api(
        cls, accqsure: "AccQsure", data: dict[str, Any]
    ) -> "DocumentType":
        if not data:
            return None
        entity = cls(
            id=data.get("entity_id"),
            name=data.get("name"),
            code=data.get("code"),
            level=data.get("level"),
            created_at=data.get("created_at"),
            updated_at=data.get("updated_at"),
        )
        entity.accqsure = accqsure
        return entity

    @property
    def accqsure(self) -> "AccQsure":
        return self._accqsure

    @accqsure.setter
    def accqsure(self, value: "AccQsure"):
        self._accqsure = value

    async def remove(self):
        await self.accqsure._query(
            f"/document/type/{self.id}",
            "DELETE",
        )

    async def update(self, **kwargs):
        resp = await self.accqsure._query(
            f"/document/type/{self.id}",
            "PUT",
            None,
            dict(**kwargs),
        )
        exclude = ["id", "accqsure"]

        for f in fields(self.__class__):
            if (
                f.name not in exclude
                and f.init
                and resp.get(f.name) is not None
            ):  # Only update init args
                setattr(self, f.name, resp.get(f.name))
        return self

    async def refresh(self):
        resp = await self.accqsure._query(
            f"/document/type/{self.id}",
            "GET",
        )
        exclude = ["id", "accqsure"]

        for f in fields(self.__class__):
            if (
                f.name not in exclude
                and f.init
                and resp.get(f.name) is not None
            ):  # Only update init args
                setattr(self, f.name, resp.get(f.name))
        return self



from __future__ import annotations
from dataclasses import dataclass, field, fields
from typing import Optional, Any, TYPE_CHECKING

from .elements import PlotElements

if TYPE_CHECKING:
    from accqsure import AccQsure


class PlotSections(object):
    def __init__(self, accqsure, plot_id):
        self.accqsure = accqsure
        self.plot_id = plot_id

    async def get(self, id_, **kwargs):

        resp = await self.accqsure._query(
            f"/plot/{self.plot_id}/section/{id_}", "GET", kwargs
        )
        return PlotSection.from_api(self.accqsure, self.plot_id, resp)

    async def list(self, limit=50, start_key=None, **kwargs):

        resp = await self.accqsure._query(
            f"/plot/{self.plot_id}/section",
            "GET",
            {"limit": limit, "start_key": start_key, **kwargs},
        )
        plot_sections = [
            PlotSection.from_api(self.accqsure, self.plot_id, plot_section)
            for plot_section in resp.get("results")
        ]
        return plot_sections, resp.get("last_key")


@dataclass
class PlotSection:
    plot_id: str
    id: str
    heading: str
    style: str
    order: int
    created_at: Optional[str] = field(default=None)
    updated_at: Optional[str] = field(default=None)
    number: Optional[str] = field(default=None)

    elements: PlotElements = field(
        init=False, repr=False, compare=False, hash=False
    )

    @classmethod
    def from_api(
        cls, accqsure: "AccQsure", plot_id: str, data: dict[str, Any]
    ) -> "PlotSection":
        if not data:
            return None
        entity = cls(
            plot_id=plot_id,
            id=data.get("entity_id"),
            heading=data.get("heading"),
            number=data.get("number"),
            style=data.get("style"),
            order=data.get("order"),
            created_at=data.get("created_at"),
            updated_at=data.get("updated_at"),
        )
        entity.accqsure = accqsure
        entity.elements = PlotElements(
            entity.accqsure, entity.plot_id, entity.id
        )
        return entity

    @property
    def accqsure(self) -> "AccQsure":
        return self._accqsure

    @accqsure.setter
    def accqsure(self, value: "AccQsure"):
        self._accqsure = value

    async def refresh(self):

        resp = await self.accqsure._query(
            f"/plot/{self.plot_id}/section/{self.id}",
            "GET",
        )
        exclude = ["id", "plot_id", "accqsure"]

        for f in fields(self.__class__):
            if (
                f.name not in exclude and f.init and resp.get(f.name) is not None
            ):  # Only update init args (skip derived like elements)
                setattr(self, f.name, resp.get(f.name))
        return self



from __future__ import annotations
from dataclasses import dataclass, field, fields
from typing import Optional, Any, TYPE_CHECKING
import logging

from accqsure.exceptions import SpecificationError
from accqsure.documents import Document

if TYPE_CHECKING:
    from accqsure import AccQsure


class Manifests(object):
    def __init__(self, accqsure):
        self.accqsure = accqsure

    async def get(self, id_, **kwargs):
        resp = await self.accqsure._query(f"/manifest/{id_}", "GET", kwargs)
        return Manifest.from_api(self.accqsure, resp)

    async def get_global(self, **kwargs):
        resp = await self.accqsure._query("/manifest/global", "GET", kwargs)
        return Manifest.from_api(self.accqsure, resp)

    async def list(
        self,
        document_type_id,
        limit=50,
        start_key=None,
        fetch_all=False,
        **kwargs,
    ):
        if fetch_all:
            resp = await self.accqsure._query_all(
                "/manifest",
                "GET",
                {
                    "document_type_id": document_type_id,
                    **kwargs,
                },
            )
            manifests = [
                Manifest.from_api(self.accqsure, manifest) for manifest in resp
            ]
            return manifests
        else:
            resp = await self.accqsure._query(
                "/manifest",
                "GET",
                {
                    "document_type_id": document_type_id,
                    "limit": limit,
                    "start_key": start_key,
                    **kwargs,
                },
            )
            manifests = [
                Manifest.from_api(self.accqsure, manifest)
                for manifest in resp.get("results")
            ]
            return manifests, resp.get("last_key")

    async def create(
        self,
        document_type_id,
        name,
        reference_document_id,
        **kwargs,
    ):

        data = dict(
            name=name,
            document_type_id=document_type_id,
            reference_document_id=reference_document_id,
            **kwargs,
        )
        payload = {k: v for k, v in data.items() if v is not None}
        logging.info("Creating Manifest %s", name)
        resp = await self.accqsure._query("/manifest", "POST", None, payload)
        manifest = Manifest.from_api(self.accqsure, resp)
        logging.info("Created Manifest %s with id %s", name, manifest.id)

        return manifest

    async def remove(self, id_, **kwargs):
        await self.accqsure._query(
            f"/manifest/{id_}", "DELETE", dict(**kwargs)
        )


@dataclass
class Manifest:
    id: str
    name: str
    document_type_id: str
    created_at: str
    updated_at: str
    global_: Optional[bool] = field(default=None)
    reference_document: Optional[Document] = field(default=None)

    @classmethod
    def from_api(
        cls, accqsure: "AccQsure", data: dict[str, Any]
    ) -> "Manifest":
        if not data:
            return None
        entity = cls(
            id=data.get("entity_id"),
            name=data.get("name"),
            document_type_id=data.get("document_type_id"),
            created_at=data.get("created_at"),
            updated_at=data.get("updated_at"),
            global_=data.get("global"),
            reference_document=Document.from_api(
                accqsure=accqsure, data=data.get("reference_document")
            )
            if data.get("reference_document")
            else None,
        )
        entity.accqsure = accqsure
        return entity

    @property
    def accqsure(self) -> "AccQsure":
        return self._accqsure

    @accqsure.setter
    def accqsure(self, value: "AccQsure"):
        self._accqsure = value

    @property
    def reference_document_id(self) -> str:
        return (
            self.reference_document.id
            if self.reference_document
            else "UNKNOWN"
        )

    @property
    def reference_document_doc_id(self) -> str:
        return (
            self.reference_document.doc_id
            if self.reference_document
            else "UNKNOWN"
        )

    async def remove(self):
        await self.accqsure._query(
            f"/manifest/{self.id}",
            "DELETE",
        )

    async def rename(self, name):
        resp = await self.accqsure._query(
            f"/manifest/{self.id}",
            "PUT",
            None,
            dict(name=name),
        )
        exclude = ["id", "accqsure"]

        for f in fields(self.__class__):
            if (
                f.name not in exclude and f.init and resp.get(f.name)
            ):  # Only update init args
                setattr(self, f.name, resp.get(f.name))
        return self

    async def refresh(self):
        resp = await self.accqsure._query(
            f"/manifest/{self.id}",
            "GET",
        )
        exclude = ["id", "accqsure", "reference_document"]

        for f in fields(self.__class__):
            if (
                f.name not in exclude and f.init and resp.get(f.name) is not None
            ):  # Only update init args
                setattr(self, f.name, resp.get(f.name))
        
        # Handle reference_document separately
        if resp.get("reference_document"):
            self.reference_document = Document.from_api(
                accqsure=self.accqsure, data=resp.get("reference_document")
            )
        elif "reference_document" in resp:
            self.reference_document = None
        
        return self

    async def get_reference_contents(self):
        if not self.reference_document:
            raise SpecificationError(
                "reference_document",
                "Reference document not found for manifest",
            )
        document_id = self.reference_document.id
        content_id = self.reference_document.content_id
        if not content_id:
            raise SpecificationError(
                "content_id", "Content not uploaded for document"
            )
        resp = await self.accqsure._query(
            f"/document/{document_id}/asset/{content_id}/manifest.json",
            "GET",
        )
        return resp

    async def get_reference_content_item(self, name):
        if not self.reference_document:
            raise SpecificationError(
                "reference_document",
                "Reference document not found for manifest",
            )
        document_id = self.reference_document.id
        content_id = self.reference_document.content_id
        if not content_id:
            raise SpecificationError(
                "content_id", "Content not uploaded for document"
            )
        resp = await self.accqsure._query(
            f"/document/{document_id}/asset/{content_id}/{name}",
            "GET",
        )
        return resp

    async def list_checks(self, limit=50, start_key=None, **kwargs):
        resp = await self.accqsure._query(
            f"/manifest/{self.id}/check",
            "GET",
            {"limit": limit, "start_key": start_key, **kwargs},
        )
        checks = [
            ManifestCheck.from_api(self.accqsure, self.id, check)
            for check in resp.get("results")
        ]
        return checks, resp.get("last_key")

    async def create_check(self, name, section, prompt, **kwargs):
        data = dict(
            name=name,
            section=section,
            prompt=prompt,
            **kwargs,
        )
        payload = {k: v for k, v in data.items() if v is not None}
        logging.info("Creating Manifest Check %s", name)
        resp = await self.accqsure._query(
            f"/manifest/{self.id}/check", "POST", None, payload
        )
        check = ManifestCheck.from_api(self.accqsure, self.id, resp)
        logging.info("Created Manifest Check %s with id %s", name, check.id)

        return check

    async def remove_check(self, check_id, **kwargs):
        await self.accqsure._query(
            f"/manifest/{self.id}/check/{check_id}", "DELETE", dict(**kwargs)
        )

    async def _set_asset(self, path, file_name, mime_type, contents):
        return await self.accqsure._query(
            f"/manifest/{self.id}/asset/{path}",
            "PUT",
            params={"file_name": file_name},
            data=contents,
            headers={"Content-Type": mime_type},
        )


@dataclass
class ManifestCheck:
    manifest_id: str
    id: str
    section: str
    name: str
    prompt: str
    critical: Optional[bool] = field(default=None)
    created_at: Optional[str] = field(default=None)
    updated_at: Optional[str] = field(default=None)

    @classmethod
    def from_api(
        cls, accqsure: "AccQsure", manifest_id: str, data: dict[str, Any]
    ) -> "ManifestCheck":
        if not data:
            return None
        entity = cls(
            manifest_id=manifest_id,
            id=data.get("entity_id"),
            section=data.get("section"),
            name=data.get("name"),
            prompt=data.get("prompt"),
            critical=data.get("critical"),
            created_at=data.get("created_at"),
            updated_at=data.get("updated_at"),
        )
        entity.accqsure = accqsure
        return entity

    @property
    def accqsure(self) -> "AccQsure":
        return self._accqsure

    @accqsure.setter
    def accqsure(self, value: "AccQsure"):
        self._accqsure = value

    async def remove(self):

        await self.accqsure._query(
            f"/manifest/{self.manifest_id}/check/{self.id}",
            "DELETE",
        )

    async def update(self, **kwargs):

        resp = await self.accqsure._query(
            f"/manifest/{self.manifest_id}/check/{self.id}",
            "PUT",
            None,
            dict(**kwargs),
        )
        exclude = ["id", "manifest_id", "accqsure"]

        for f in fields(self.__class__):
            if (
                f.name not in exclude and f.init and resp.get(f.name)
            ):  # Only update init args
                setattr(self, f.name, resp.get(f.name))
        return self

    async def refresh(self):

        resp = await self.accqsure._query(
            f"/manifest/{self.manifest_id}/check/{self.id}",
            "GET",
        )
        exclude = ["id", "manifest_id", "accqsure"]

        for f in fields(self.__class__):
            if (
                f.name not in exclude and f.init and resp.get(f.name)
            ):  # Only update init args
                setattr(self, f.name, resp.get(f.name))
        return self



from __future__ import annotations
from dataclasses import dataclass, field, fields
from typing import Optional, Any, TYPE_CHECKING
import logging

from accqsure.exceptions import SpecificationError

if TYPE_CHECKING:
    from accqsure import AccQsure


class PlotMarkers(object):
    def __init__(
        self,
        accqsure,
        plot_id,
        plot_waypoint_id,
    ):
        self.accqsure = accqsure
        self.plot_id = plot_id
        self.waypoint_id = plot_waypoint_id

    async def get(self, id_, **kwargs):

        resp = await self.accqsure._query(
            f"/plot/{self.plot_id}/waypoint/{self.waypoint_id}/marker/{id_}",
            "GET",
            kwargs,
        )
        return PlotMarker.from_api(
            self.accqsure, self.plot_id, self.waypoint_id, resp
        )

    async def list(self, limit=50, start_key=None, **kwargs):

        resp = await self.accqsure._query(
            f"/plot/{self.plot_id}/waypoint/{self.waypoint_id}/marker",
            "GET",
            {"limit": limit, "start_key": start_key, **kwargs},
        )
        plot_markers = [
            PlotMarker.from_api(
                self.accqsure, self.plot_id, self.waypoint_id, plot_marker
            )
            for plot_marker in resp.get("results")
        ]
        return plot_markers, resp.get("last_key")

    async def create(
        self,
        name,
        contents,
        **kwargs,
    ):

        data = dict(
            name=name,
            contents=contents,
            **kwargs,
        )
        payload = {k: v for k, v in data.items() if v is not None}
        logging.info("Creating Plot Marker %s", name)

        resp = await self.accqsure._query(
            f"/plot/{self.plot_id}/waypoint/{self.waypoint_id}/marker",
            "POST",
            None,
            payload,
        )
        plot_marker = PlotMarker.from_api(
            self.accqsure, self.plot_id, self.waypoint_id, resp
        )
        logging.info("Created Plot Marker %s with id %s", name, plot_marker.id)

        return plot_marker

    async def remove(self, id_, **kwargs):

        await self.accqsure._query(f"/plot/{id_}", "DELETE", {**kwargs})


@dataclass
class PlotMarker:
    plot_id: str
    waypoint_id: str
    id: str
    name: str
    status: str
    created_at: Optional[str] = field(default=None)
    updated_at: Optional[str] = field(default=None)
    content_id: Optional[str] = field(default=None)

    @classmethod
    def from_api(
        cls,
        accqsure: "AccQsure",
        plot_id: str,
        waypoint_id: str,
        data: dict[str, Any],
    ) -> "PlotMarker":
        if not data:
            return None
        entity = cls(
            plot_id=plot_id,
            waypoint_id=waypoint_id,
            id=data.get("entity_id"),
            name=data.get("name"),
            status=data.get("status"),
            content_id=data.get("content_id"),
            created_at=data.get("created_at"),
            updated_at=data.get("updated_at"),
        )
        entity.accqsure = accqsure
        return entity

    @property
    def accqsure(self) -> "AccQsure":
        return self._accqsure

    @accqsure.setter
    def accqsure(self, value: "AccQsure"):
        self._accqsure = value

    async def remove(self):

        await self.accqsure._query(
            f"/plot/{self.plot_id}/waypoint/{self.waypoint_id}/marker/{self.id}",
            "DELETE",
        )

    async def rename(self, name):

        resp = await self.accqsure._query(
            f"/plot/{self.plot_id}/waypoint/{self.waypoint_id}/marker/{self.id}",
            "PUT",
            None,
            dict(name=name),
        )
        exclude = ["id", "plot_id", "waypoint_id", "accqsure"]

        for f in fields(self.__class__):
            if (
                f.name not in exclude and f.init and resp.get(f.name) is not None
            ):  # Only update init args
                setattr(self, f.name, resp.get(f.name))
        return self

    async def refresh(self):

        resp = await self.accqsure._query(
            f"/plot/{self.plot_id}/waypoint/{self.waypoint_id}/marker/{self.id}",
            "GET",
        )
        exclude = ["id", "plot_id", "waypoint_id", "accqsure"]

        for f in fields(self.__class__):
            if (
                f.name not in exclude and f.init and resp.get(f.name) is not None
            ):  # Only update init args
                setattr(self, f.name, resp.get(f.name))
        return self

    async def _set_asset(self, path, file_name, mime_type, contents):
        return await self.accqsure._query(
            f"/plot/{self.plot_id}/waypoint/{self.waypoint_id}/marker/{self.id}/asset/{path}",
            "PUT",
            params={"file_name": file_name},
            data=contents,
            headers={"Content-Type": mime_type},
        )

    async def get_contents(self):
        if not self.content_id:
            raise SpecificationError(
                "content_id", "Content not ready for plot marker"
            )

        resp = await self.accqsure._query(
            f"/plot/{self.plot_id}/waypoint/{self.waypoint_id}/marker/{self.id}/asset/manifest.json",
            "GET",
        )
        return resp

    async def get_content_item(self, name):
        if not self.content_id:
            raise SpecificationError(
                "content_id", "Content not ready for plot marker"
            )

        return await self.accqsure._query(
            f"/plot/{self.plot_id}/waypoint/{self.waypoint_id}/marker/{self.id}/asset/{name}",
            "GET",
        )

    async def _set_content_item(self, name, file_name, mime_type, contents):
        if not self.content_id:
            raise SpecificationError(
                "content_id", "Content not ready for plot marker"
            )
        return await self._set_asset(name, file_name, mime_type, contents)



from __future__ import annotations
from dataclasses import dataclass, field, fields
from typing import Optional, Any, TYPE_CHECKING

if TYPE_CHECKING:
    from accqsure import AccQsure


class PlotElements(object):
    def __init__(
        self,
        accqsure,
        plot_id,
        plot_section_id,
    ):
        self.accqsure = accqsure
        self.plot_id = plot_id
        self.section_id = plot_section_id

    async def get(self, id_, **kwargs):

        resp = await self.accqsure._query(
            f"/plot/{self.plot_id}/section/{self.section_id}/element/{id_}",
            "GET",
            kwargs,
        )
        return PlotElement.from_api(
            self.accqsure, self.plot_id, self.section_id, resp
        )

    async def list(self, limit=50, start_key=None, **kwargs):

        resp = await self.accqsure._query(
            f"/plot/{self.plot_id}/section/{self.section_id}/element",
            "GET",
            {"limit": limit, "start_key": start_key, **kwargs},
        )
        plot_elements = [
            PlotElement.from_api(
                self.accqsure, self.plot_id, self.section_id, plot_element
            )
            for plot_element in resp.get("results")
        ]
        return plot_elements, resp.get("last_key")


@dataclass
class PlotElement:
    plot_id: str
    section_id: str
    id: str
    order: int
    type: str
    status: str
    created_at: Optional[str] = field(default=None)
    updated_at: Optional[str] = field(default=None)
    content: Optional[str] = field(default=None)

    @classmethod
    def from_api(
        cls,
        accqsure: "AccQsure",
        plot_id: str,
        section_id: str,
        data: dict[str, Any],
    ) -> "PlotElement":
        if not data:
            return None
        entity = cls(
            plot_id=plot_id,
            section_id=section_id,
            id=data.get("entity_id"),
            order=data.get("order"),
            type=data.get("type"),
            status=data.get("status"),
            content=data.get("content"),
            created_at=data.get("created_at"),
            updated_at=data.get("updated_at"),
        )
        entity.accqsure = accqsure
        return entity

    @property
    def accqsure(self) -> "AccQsure":
        return self._accqsure

    @accqsure.setter
    def accqsure(self, value: "AccQsure"):
        self._accqsure = value

    async def refresh(self):

        resp = await self.accqsure._query(
            f"/plot/{self.plot_id}/section/{self.section_id}/element/{self.id}",
            "GET",
        )
        exclude = ["id", "plot_id", "section_id", "accqsure"]

        for f in fields(self.__class__):
            if (
                f.name not in exclude and f.init and resp.get(f.name) is not None
            ):  # Only update init args
                setattr(self, f.name, resp.get(f.name))
        return self
